import time

from typing import Any, Dict, Type
from uuid import UUID

from clickhouse_connect import get_async_client
from clickhouse_connect.driver.asyncclient import AsyncClient
from ulid import ULID

from .base import BaseDBODriver
from ...dbo import ClickhouseModelDBOType
from ...settings import ClickhouseSettings


class ClickhouseDBODriver(BaseDBODriver):
    def __init__(self, settings: ClickhouseSettings):
        self.settings = settings

    async def connect(self) -> AsyncClient:
        return await get_async_client(
            username=self.settings.dsn.username,
            host=self.settings.dsn.host,
            port=int(self.settings.dsn.port),
            password=self.settings.dsn.password,
            database=self.settings.dsn.path.replace("/", ""),
        )
    
    def adapt_value(self, value: Any) -> Any:
        if isinstance(value, UUID):
            return str(value)
        else:
            return value

    async def create_model(self, dbo_cls: ClickhouseModelDBOType):
        client = await self.connect()
        table_list_res = await client.query("SHOW TABLES;")
        table_list = [ item[0] for item in table_list_res.result_rows ]
        if dbo_cls.__tablename__ not in table_list:
            await client.command(dbo_cls.get_create_query())

    async def init_model(self, dbo_cls: ClickhouseModelDBOType):
        pass

    async def rename_model(self, old_name: str, new_name: str):
        try:
            client = await self.connect()
            result = await client.command(f"RENAME TABLE {old_name} TO {new_name};")

            return result
        except:
            return None

    async def get_models_list(self) -> list[str]:
        client = await self.connect()
        res = await client.query("SHOW TABLES;")
        return [ item[0] for item in res.result_rows ]

    async def save(self, dbo: ClickhouseModelDBOType, **extra):
        if dbo.id:
            raise Exception("Clickhouse driver does not support update operations")
        
        dbo.id = ULID().to_uuid4()
        dbo.ts = time.time_ns()

        dbo_data = dbo.model_dump()
        values = [ self.adapt_value(item) for item in list(dbo_data.values()) ]
        client = await self.connect()
        await client.insert(dbo.__tablename__, [values])

        return dbo

    async def get(self, dbo_cls: ClickhouseModelDBOType, attr: str, value: Any) -> ClickhouseModelDBOType | None:
        request = f"select * from {dbo_cls.__tablename__} where {attr} = '{value}' limit 1;"

        client = await self.connect()
        res = await client.query(request)

        if len(res.result_rows) < 1:
            return None
        
        res = res.result_rows[0]

        res_data = {}
        for i in range(len(res)):
            value = res[i]
            key = dbo_cls.get_fields()[i]
            res_data[key] = value

        return dbo_cls.model_validate(res_data)

    async def get_list(
        self,
        dbo_cls: ClickhouseModelDBOType,
        attrs: Dict[str, Any] | None = None,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0,
    ) -> list[ClickhouseModelDBOType]:
        if order_by not in dbo_cls.get_fields():
            order_by = "id"

        conditions = []
        for key, attr in attrs.items():
            conditions.append(f"{key} = '{attr}'")

        request = f"select * from {dbo_cls.__tablename__} where {' and '.join(conditions)} order by {order_by} {order_direction} limit {limit};"
        result = await self.query(dbo_cls, request)
        return result

    async def delete(self, dbo: ClickhouseModelDBOType):
        raise Exception("Clickhouse driver does not support delete operations")

    async def get_count(
        self,
        dbo_cls: Type[ClickhouseModelDBOType],
        attrs: Dict[str, Any] | None = None
    ) -> int:
        conditions = []
        for key, attr in attrs.items():
            conditions.append(f"{key} = '{attr}'")

        request = f"select count() from {dbo_cls.__tablename__} where {' and '.join(conditions)};"
        client = await self.connect()
        res = await client.query(request)
        return int(res.result_rows[0][0])

    async def raw_execute(self, request: str): # pragma: no cover
        client = await self.connect()
        await client.command(request)

    async def raw_query(self, request: str) -> list[ClickhouseModelDBOType]: # pragma: no cover
        client = await self.connect()
        return await client.query(request)
    
    async def query(self, dbo_cls: ClickhouseModelDBOType, request: str) -> list[ClickhouseModelDBOType]:
        result = []

        client = await self.connect()
        res = await client.query(request)

        for row in res.result_rows:
            row_data = {}
            for i in range(len(row)):
                value = row[i]
                key = dbo_cls.get_fields()[i]
                row_data[key] = value
            result.append(dbo_cls.model_validate(row_data))

        return result



from pydantic import BaseModel


class CommunicationsSettings(BaseModel):
    email_actions_topic: str = "email_actions"
    email_sender: str | None = None

    smtp_host: str | None = None
    smtp_port: str | None = None
    smtp_username: str | None = None
    smtp_password: str | None = None



from typing import Self


class BaseMigration:
    next: Self | None = None

    @property
    def migration_id(self):
        return self.__class__.__name__

    async def migrate(self) -> str:
        print(f"Run migration: [{self.migration_id}]")
        await self.upgrade()
        print(f"End migration: [{self.migration_id}]")

        if self.next:
            return await self.next.migrate()
        
        return self.migration_id

    def find_last_migration(self, migration_id: str) -> Self | None:
        if migration_id == self.migration_id:
            return self

        if self.next:
            return self.next.find_last_migration(migration_id)

        return None

    async def upgrade(self):
        pass



from .rethinkdb import RethinkDBBaseDBO, ModelDBOType as RethinkDBModelDBOType
from .clickhouse import ClickhouseBaseDBO, ModelDBOType as ClickhouseModelDBOType


__all__ = [
    "RethinkDBBaseDBO", "RethinkDBModelDBOType",
    "ClickhouseBaseDBO", "ClickhouseModelDBOType",
]



from pydantic import BaseModel


class UsersSettings(BaseModel):
    debug_admin_password: str | None = None
    auth_token_length: int = 32
    auth_access_token_expire: int = 7*24*60*60

    auth_magic_link_token_length: int = 16
    auth_magic_link_token_ttl: int = 30

    user_account_id_alphabet: str = "ABCDEFGHJKLMNPRTUVWXYZ0123456789"
    user_account_id_length: int = 10



from pydantic import BaseModel


class FilesSettings(BaseModel):
    s3_access_key: str | None = None
    s3_secret_key: str | None = None
    s3_host: str | None = None
    s3_region: str | None = None
    s3_bucket: str | None = None



from typing import TypeVar, Union, Annotated, get_origin, get_args
from uuid import UUID

from pydantic import UUID4
from pydantic.types import UuidVersion
from ulid import ULID

from .base import BaseDBO


class ClickhouseBaseDBO(BaseDBO):
    __tablename__: str = "objects"
    __engine__: str = "MergeTree"
    __index_granularity__: int = 8192

    id: UUID4 | None = None
    ts: int = -1

    @classmethod
    def get_fields(cls) -> list[str]:
        return [ field_name for field_name, _ in cls.model_fields.items() ]
    
    @classmethod
    def get_create_query(cls) -> str:
        fields = [
            "`id` UUID",
            "`ts` UInt64",
        ]

        for field_name, field_info in cls.model_fields.items():
            if field_name not in ["id", "ts"]:
                is_optional = False
                field_type = "String"
                annotation_args = get_args(field_info.annotation)
                annotation_origin = get_origin(field_info.annotation)

                if annotation_origin == Union:
                    for annotation_arg in annotation_args:
                        if annotation_arg == None.__class__:
                            is_optional = True
                        else:
                            if get_origin(annotation_arg) == Annotated:
                                annotated_args = get_args(annotation_arg)
                                if UUID in annotated_args:
                                    field_type = "UUID"

                if field_info.annotation == int:
                    field_type = "Int64"
                elif field_info.annotation == float:
                    field_type = "Float64"
                elif field_info.annotation in [UUID, UUID4, ULID]:
                    field_type = "UUID"

                if is_optional:
                    result_field_type = f"Nullable({field_type})"
                else:
                    result_field_type = field_type

                fields.append(f"`{field_name}` {result_field_type}")

        query = f"""
            CREATE TABLE IF NOT EXISTS {cls.__tablename__}
            ({",".join(fields)})
            ENGINE = {cls.__engine__}
            PRIMARY KEY id
            ORDER BY id
            SETTINGS index_granularity = {cls.__index_granularity__}
            COMMENT 'migration:1';
        """

        return query


ModelDBOType = TypeVar('ModelDBOType', bound=ClickhouseBaseDBO)



from .rethinkdb import RethinkDBSettings
from .clickhouse import ClickhouseSettings
from .kafka import KafkaSettings
from .communications import CommunicationsSettings
from .users import UsersSettings
from .files import FilesSettings


__all__ = [
    "RethinkDBSettings",
    "ClickhouseSettings",
    "KafkaSettings",
    "CommunicationsSettings",
    "UsersSettings",
    "FilesSettings",
]
